@@ -9,35 +9,30 @@
  * Message queues
  */
 
-struct queue_node {
-	message message;
-	struct queue_node* next;
-};
-
-static void node_unshift(queue_node** position, queue_node* new_node) {
+static void node_unshift(message** position, message* new_node) {
 	new_node->next = *position;
 	*position = new_node;
 }
 
-static queue_node* node_shift(queue_node** position) {
-	queue_node* ret = *position;
+static const message* node_shift(message** position) {
+	message* ret = *position;
 	*position = (*position)->next;
+	ret->next = NULL;
 	return ret;
 }
 
-static void node_push(queue_node** end, queue_node* new_node) {
-	queue_node** cur = end;
+static void node_push(message** end, message* new_node) {
+	message** cur = end;
 	while(*cur)
 		cur = &(*cur)->next;
 	*end = *cur = new_node;
 	new_node->next = NULL;
 }
 
-static void S_node_destroy(pTHX_ struct queue_node** current) {
+static void S_node_destroy(pTHX_ message** current) {
 	while (*current != NULL) {
-		struct queue_node** next = &(*current)->next;
-		message_destroy(&(*current)->message);
-		PerlMemShared_free(*current);
+		message** next = &(*current)->next;
+		destroy_message(*current);
 		*current = NULL;
 		current = next;
 	}
@@ -50,22 +45,13 @@ void queue_init(message_queue* queue) {
 	COND_INIT(&queue->condvar);
 }
 
-void S_queue_enqueue(pTHX_ message_queue* queue, message* message_, perl_mutex* external_lock) {
-	queue_node* new_entry;
+void S_queue_enqueue(pTHX_ message_queue* queue, const message* message_, perl_mutex* external_lock) {
+	message* new_entry;
 	MUTEX_LOCK(&queue->mutex);
 	if (external_lock)
 		MUTEX_UNLOCK(external_lock);
 
-	if (queue->reserve)
-		new_entry = node_shift(&queue->reserve);
-	else
-		new_entry = PerlMemShared_calloc(1, sizeof *new_entry);
-
-	Copy(message_, &new_entry->message, 1, message);
-	Zero(message_, 1, message);
-	new_entry->next = NULL;
-
-	node_push(&queue->back, new_entry);
+	node_push(&queue->back, (message*)message_);
 	if (queue->front == NULL)
 		queue->front = queue->back;
 
@@ -73,17 +59,16 @@ void S_queue_enqueue(pTHX_ message_queue* queue, message* message_, perl_mutex*
 	MUTEX_UNLOCK(&queue->mutex);
 }
 
-static void queue_shift(message_queue* queue, message* input) {
-	queue_node* front = node_shift(&queue->front);
-	Copy(&front->message, input, 1, message);
-	Zero(&front->message, 1, message);
-	node_unshift(&queue->reserve, front);
+static const message* queue_shift(message_queue* queue) {
+	const message* ret = node_shift(&queue->front);
 
 	if (queue->front == NULL)
 		queue->back = NULL;
+	return ret;
 }
 
-void S_queue_dequeue(pTHX_ message_queue* queue, message* input, perl_mutex* external_lock) {
+const message* S_queue_dequeue(pTHX_ message_queue* queue, perl_mutex* external_lock) {
+	const message* ret;
 	MUTEX_LOCK(&queue->mutex);
 	if (external_lock)
 		MUTEX_UNLOCK(external_lock);
@@ -91,32 +76,32 @@ void S_queue_dequeue(pTHX_ message_queue* queue, message* input, perl_mutex* ext
 	while (!queue->front)
 		COND_WAIT(&queue->condvar, &queue->mutex);
 
-	queue_shift(queue, input);
-
+	ret = queue_shift(queue);
 	MUTEX_UNLOCK(&queue->mutex);
+
+	return ret;
 }
 
-bool S_queue_dequeue_nb(pTHX_ message_queue* queue, message* input, perl_mutex* external_lock) {
+const message* S_queue_dequeue_nb(pTHX_ message_queue* queue, perl_mutex* external_lock) {
 	MUTEX_LOCK(&queue->mutex);
 	if (external_lock)
 		MUTEX_UNLOCK(external_lock);
 
 	if (queue->front) {
-		queue_shift(queue, input);
+		const message* ret = queue_shift(queue);
 
 		MUTEX_UNLOCK(&queue->mutex);
-		return TRUE;
+		return ret;
 	}
 	else {
 		MUTEX_UNLOCK(&queue->mutex);
-		return FALSE;
+		return NULL;
 	}
 }
 
 void S_queue_destroy(pTHX_ message_queue* queue) {
 	MUTEX_LOCK(&queue->mutex);
 	node_destroy(&queue->front);
-	node_destroy(&queue->reserve);
 	COND_DESTROY(&queue->condvar);
 	MUTEX_UNLOCK(&queue->mutex);
 	MUTEX_DESTROY(&queue->mutex);
